;
;
; *****************************************************************
; *                PureBasic COM Object Framework                 *
; *****************************************************************
; *                                                               *
; *              02/25/2006  by Timo 'fr34k' Harter               * 
; *                    purebasic@freak1.de                        *
; *                                                               *
; *  Please consult the 'Readme.txt' that came with this archive  *
; *  on the use of these macros/commands. This framework may      *
; *  be used under the terms of the MIT license displayed below.  *
; *                                                               *
; *****************************************************************
;
;
; Copyright (c) 2006 Timo Harter
;
; Permission is hereby granted, free of charge, to any person obtaining a copy 
; of this software and associated documentation files (the "Software"), to deal 
; in the Software without restriction, including without limitation the rights 
; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
; of the Software, and to permit persons to whom the Software is furnished to do 
; so, subject to the following conditions:
; 
; The above copyright notice and this permission notice shall be included in 
; all copies or substantial portions of the Software.
; 
; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
; INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
; PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
; OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
; SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
;
;
;

; set the frameworks debugging flag
#__COM_DEBUGGING_ENABLED = 1

; Available Information depending on the DebugLevel:
Enumeration
  #COM_DEBUG_Critical         ; invalid access to an interface
  #COM_DEBUG_UnknownInterface ; somebody quierying an unknown Interface
  #COM_DEBUG_ObjectStats      ; Object Refcounts, Addref/Release/QueryInterface calls
  #COM_DEBUG_MethodEntry      ; calls to each method
  #COM_DEBUG_MethodLeave      ; returns from each method (including return value)
  #COM_DEBUG_FormatResult     ; displays constant name and error message for the method result
  #COM_DEBUG_All              ; highest debug level -> all output (default)
EndEnumeration

DebugLevel #COM_DEBUG_All

Macro EnableCOMDeadTracking
  #__COM_DEADTRACKING_ENABLED = 1
EndMacro

; allow the use of OutputDebugString_() if the debugger is off
; (for example to debug a COM dll)
;
CompilerIf #PB_Compiler_Debugger
  Macro __COM_DEBUG(text, level=0)
    Debug "[COM] "+text, level
  EndMacro
CompilerElse
  Procedure __COM_DEBUG(Text$, level=0)
    Text$ = "[COM] " + Text$
    OutputDebugString_(@Text$)
  EndProcedure
CompilerEndIf

XIncludeFile #PB_Compiler_Home+"COMFramework\GuidDebugList.pb"
XIncludeFile #PB_Compiler_Home+"COMFramework\ErrorList.pb"

Structure __COM_DEBUG_ObjectList
  ObjectPointer.l
  IUnknownVtbl.l
  NbInterfaces.l
  *RefCountPtr.LONG
  ObjectNameString$
  InterfaceNameString$[9]
EndStructure

Procedure __COM_DEBUG_ObjectListInit()
  Static Init  
  If Init = 0
    Global NewList __COM_DEBUG_ObjectList.__COM_DEBUG_ObjectList()

    CompilerIf #PB_Compiler_Thread
      Global __COM_DEBUG_ObjectListMutex = CreateMutex()
    CompilerEndIf
    
    Init = 1
  EndIf   
EndProcedure

Procedure DisplayCOMObjectList()

  CompilerIf #PB_Compiler_Thread
    LockMutex(__COM_DEBUG_ObjectListMutex)
  CompilerEndIf
  
  __COM_DEBUG("Listing existing COM objects:")
  
  If CountList(__COM_DEBUG_ObjectList()) = 0
    __COM_DEBUG("  There are no objects.")
  Else
  
    ForEach __COM_DEBUG_ObjectList()
      __COM_DEBUG("  " + __COM_DEBUG_ObjectList()\ObjectNameString$ + " - ObjectRefCount: "+Str(__COM_DEBUG_ObjectList()\RefCountPtr\l))
    Next __COM_DEBUG_ObjectList()

  EndIf
  
  __COM_DEBUG("Listing complete")
    
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(__COM_DEBUG_ObjectListMutex)
  CompilerEndIf    

EndProcedure

Procedure.s NameFromGUID(*guid, Type$="GUID")
  Protected *pointer = ?__COM_DEBUG_GUIDValues
  Protected Name$    = "Unknown "+Type$+" Value"
  Protected index, i, GetName$
  
  For index = 1 To #__COM_DEBUG_GUIDCount
    If CompareGUID(*pointer, *guid)
      Restore __COM_DEBUG_GUIDNames
      For i = 1 To index
        Read GetName$
      Next i
      
      If Left(GetName$, Len(Type$)+1) = Type$+"_"
        Name$ = GetName$
        Break
      EndIf
    EndIf
    
    *pointer + SizeOf(GUID)
  Next index
  
  ProcedureReturn Name$   
EndProcedure

Macro NameFromIID(iid)
  NameFromGUID(iid, "IID")
EndMacro

Macro NameFromCLSID(clsid)
  NameFromGUID(clsid, "CLSID")
EndMacro

Procedure.l GUIDFromName(Name$)
  Static EmptyGUID.GUID
  Protected *Result = @EmptyGUID, index, StoredName$

  Restore __COM_DEBUG_GUIDNames
  
  For index = 1 To #__COM_DEBUG_GUIDCount
    Read StoredName$
    If UCase(StoredName$) = UCase(Name$)
      *Result = ?__COM_DEBUG_GUIDValues + (index-1)*SizeOf(GUID)           
      Break
    EndIf       
  Next index
  
  ProcedureReturn *Result
EndProcedure

Macro IIDFromName
  GUIDFromName
EndMacro

Macro CLSIDFromName
  GUIDFromName
EndMacro

Procedure.s GetCOMErrorConstant(ErrorCode)
  Protected i, Code, Constant$, Message$
  
  Restore __COM_ErrorCodes
  For i = 1 To #__COM_ErrorCodeCount
    Read Code
    Read Constant$
    Read Message$
    
    If Code = ErrorCode
      ProcedureReturn Constant$
    EndIf
  Next i
  
  ProcedureReturn "<Unknown>"
EndProcedure

Procedure.s GetCOMErrorMessage(ErrorCode)
  Protected i, Code, Constant$, Message$
  
  Restore __COM_ErrorCodes
  For i = 1 To #__COM_ErrorCodeCount
    Read Code
    Read Constant$
    Read Message$
    
    If Code = ErrorCode
      ProcedureReturn Message$
    EndIf
  Next i
  
  ProcedureReturn ""
EndProcedure


Macro __COM_DEBUG_IUnknownInterface(Text1, Text2, Level, Object, I1, I2, I3, I4, I5, I6, I7, I8, I9)
  If *_THIS = @*THIS\IUnknown
    __COM_DEBUG(Text1+__COM_STRING(Object)+"::IUnknown"+Text2, Level)      
  ElseIf *_THIS = @*THIS\I1
    __COM_DEBUG(Text1+__COM_STRING(Object)+"::"+__COM_STRING(I1)+Text2, Level) 
    
  CompilerIf Defined(__COM_INTERFACE_#Object#_#I2#_DEFINED__, #PB_Constant)
    ElseIf *_THIS = @*THIS\I2
      __COM_DEBUG(Text1+__COM_STRING(Object)+"::"+__COM_STRING(I2)+Text2, Level) 
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#Object#_#I3#_DEFINED__, #PB_Constant)
    ElseIf *_THIS = @*THIS\I3
      __COM_DEBUG(Text1+__COM_STRING(Object)+"::"+__COM_STRING(I3)+Text2, Level) 
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#Object#_#I4#_DEFINED__, #PB_Constant)
    ElseIf *_THIS = @*THIS\I4
      __COM_DEBUG(Text1+__COM_STRING(Object)+"::"+__COM_STRING(I4)+Text2, Level) 
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#Object#_#I5#_DEFINED__, #PB_Constant)
    ElseIf *_THIS = @*THIS\I5
      __COM_DEBUG(Text1+__COM_STRING(Object)+"::"+__COM_STRING(I5)+Text2, Level) 
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#Object#_#I6#_DEFINED__, #PB_Constant)
    ElseIf *_THIS = @*THIS\I6
      __COM_DEBUG(Text1+__COM_STRING(Object)+"::"+__COM_STRING(I6)+Text2, Level) 
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#Object#_#I7#_DEFINED__, #PB_Constant)
    ElseIf *_THIS = @*THIS\I7
      __COM_DEBUG(Text1+__COM_STRING(Object)+"::"+__COM_STRING(I7)+Text2, Level) 
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#Object#_#I8#_DEFINED__, #PB_Constant)
    ElseIf *_THIS = @*THIS\I8
      __COM_DEBUG(Text1+__COM_STRING(Object)+"::"+__COM_STRING(I8)+Text2, Level) 
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#Object#_#I9#_DEFINED__, #PB_Constant)
    ElseIf *_THIS = @*THIS\I9
      __COM_DEBUG(Text1+__COM_STRING(Object)+"::"+__COM_STRING(I9)+Text2, Level) 
  CompilerEndIf
  
  Else
    __COM_DEBUG(Text1+__COM_STRING(Object)+"::<Unknown Interface>"+Text2, Level)     
  EndIf
EndMacro


Macro __COM_DEBUG_RegisterObject(THIS, ObjectName, I1,I2,I3,I4,I5,I6,I7,I8,I9) 
  __COM_DEBUG_ObjectListInit()
  
  CompilerIf #PB_Compiler_Thread
    LockMutex(__COM_DEBUG_ObjectListMutex)
  CompilerEndIf  
  
  AddElement(__COM_DEBUG_ObjectList())
  
  __COM_DEBUG_ObjectList()\ObjectPointer      = THIS
  __COM_DEBUG_ObjectList()\ObjectNameString$  = __COM_STRING(ObjectName)
  __COM_DEBUG_ObjectList()\IUnknownVtbl       = ?__COM_#ObjectName#_IUnknown_Vtbl
  __COM_DEBUG_ObjectList()\RefCountPtr        = THIS + OffsetOf(ObjectName\__COM_ObjectRefCount) 
  
  CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I1#_EXTENDS__, #PB_Constant) = 0
    __COM_DEBUG_ObjectList()\InterfaceNameString$[0] = __COM_STRING(I1)
    __COM_DEBUG_ObjectList()\NbInterfaces       = 1
  CompilerEndIf
  
  CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I2#_DEFINED__, #PB_Constant)
    CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I2#_EXTENDS__, #PB_Constant) = 0
      __COM_DEBUG_ObjectList()\InterfaceNameString$[1] = __COM_STRING(I2)
      __COM_DEBUG_ObjectList()\NbInterfaces       = 2  
    CompilerEndIf
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I3#_DEFINED__, #PB_Constant)
    CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I3#_EXTENDS__, #PB_Constant) = 0
      __COM_DEBUG_ObjectList()\InterfaceNameString$[2] = __COM_STRING(I3)
      __COM_DEBUG_ObjectList()\NbInterfaces       = 3  
    CompilerEndIf
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I4#_DEFINED__, #PB_Constant)
    CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I4#_EXTENDS__, #PB_Constant) = 0
      __COM_DEBUG_ObjectList()\InterfaceNameString$[3] = __COM_STRING(I4)
      __COM_DEBUG_ObjectList()\NbInterfaces       = 4  
    CompilerEndIf
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I5#_DEFINED__, #PB_Constant)
    CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I5#_EXTENDS__, #PB_Constant) = 0
      __COM_DEBUG_ObjectList()\InterfaceNameString$[4] = __COM_STRING(I5)
      __COM_DEBUG_ObjectList()\NbInterfaces       = 5  
    CompilerEndIf
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I6#_DEFINED__, #PB_Constant)
    CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I6#_EXTENDS__, #PB_Constant) = 0
      __COM_DEBUG_ObjectList()\InterfaceNameString$[5] = __COM_STRING(I6)
      __COM_DEBUG_ObjectList()\NbInterfaces       = 6  
    CompilerEndIf
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I7#_DEFINED__, #PB_Constant)
    CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I7#_EXTENDS__, #PB_Constant) = 0
      __COM_DEBUG_ObjectList()\InterfaceNameString$[6] = __COM_STRING(I7)
      __COM_DEBUG_ObjectList()\NbInterfaces       = 7  
    CompilerEndIf
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I8#_DEFINED__, #PB_Constant)
    CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I8#_EXTENDS__, #PB_Constant) = 0
      __COM_DEBUG_ObjectList()\InterfaceNameString$[7] = __COM_STRING(I8)
      __COM_DEBUG_ObjectList()\NbInterfaces       = 8  
    CompilerEndIf
  CompilerEndIf
  CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I9#_DEFINED__, #PB_Constant)
    CompilerIf Defined(__COM_INTERFACE_#ObjectName#_#I9#_EXTENDS__, #PB_Constant) = 0
      __COM_DEBUG_ObjectList()\InterfaceNameString$[8] = __COM_STRING(I9)
      __COM_DEBUG_ObjectList()\NbInterfaces       = 9  
    CompilerEndIf
  CompilerEndIf  
  
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(__COM_DEBUG_ObjectListMutex)
  CompilerEndIf    
  
EndMacro

Macro __COM_DEBUG_UnregisterObject(THIS, ObjectName)
  Protected __COM_Interface.l

  CompilerIf #PB_Compiler_Thread
    LockMutex(__COM_DEBUG_ObjectListMutex)
  CompilerEndIf  

  ForEach __COM_DEBUG_ObjectList()
    If __COM_DEBUG_ObjectList()\ObjectPointer = THIS

      DeleteElement(__COM_DEBUG_ObjectList())
      Break
      
    EndIf
  Next __COM_DEBUG_ObjectList()     
  
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(__COM_DEBUG_ObjectListMutex)
  CompilerEndIf    
  
EndMacro

Macro __COM_DEBUG_IUnknownExtend
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()
  Data.l @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod(), @__COM_CrashTraceMethod()  
EndMacro

Macro __COM_VTABLE_DebugExtend(InterfaceName)
  Structure __COM_#InterfaceName#_Vtbl_DEBUG Extends __COM_#InterfaceName#_Vtbl
    __COM_CrashTraceMethod.l[100]
    __COM_RealMethods.__COM_#InterfaceName#_Vtbl
  EndStructure
EndMacro

; this is created for each object separately through
; a macro. The use of this is by defining it only inside the ImplementObject
; macro, a change in DebugLevel can have an effect for each object separately
;
Macro __COM_DEBUG_ImplementPrintResult(ObjectName)
  Procedure __COM_DEBUG_#ObjectName#_PrintResult(Value, Name$)
    Protected Message$
    __COM_DEBUG("RETURNED: "+Name$+"() ReturnValue: "+Str(Value)+"  ( $"+Hex(Value)+" )", #COM_DEBUG_MethodLeave)  
    
    If Value = 0 ; #S_OK is the most common used name for success in COM (not ERROR_SUCCESS, which is otherwise shown)
      __COM_DEBUG("RETURNVALUE: Equals Constant: S_OK", #COM_DEBUG_FormatResult)
      __COM_DEBUG("RETURNVALUE: Equals Message: The operation completed successfully.", #COM_DEBUG_FormatResult)  
    ElseIf Value > 0
      __COM_DEBUG("RETURNVALUE: Equals Constant: <Unknown>", #COM_DEBUG_FormatResult)
      __COM_DEBUG("RETURNVALUE: Equals Message: Positive HRESULT generally implies success.", #COM_DEBUG_FormatResult)      
    Else
      __COM_DEBUG("RETURNVALUE: Equals Constant: "+GetCOMErrorConstant(Value), #COM_DEBUG_FormatResult)
      Message$ = GetCOMErrorMessage(Value)
      If Message$ <> ""
        __COM_DEBUG("RETURNVALUE: Equals Message: "+Message$, #COM_DEBUG_FormatResult)  
      EndIf      
    EndIf          
  EndProcedure    
  
  ; this is for access to the procedure label from direct asm code
  DataSection
    !__COM_DEBUG_#ObjectName#_PrintResult#__COM_COLON
    Data.l @__COM_DEBUG_#ObjectName#_PrintResult()
  EndDataSection
  
  ; implement the IsDead check on a per-object basis as well, as it may be
  ; disabled inbeween
  ;
  CompilerIf Defined(__COM_DEADTRACKING_ENABLED, #PB_Constant) 
    Procedure __COM_DEBUG_#ObjectName#_CheckDead(*THIS.ObjectName, Name$) 
      While *THIS\IUnknown <> ?__COM_#ObjectName#_IUnknown_Vtbl: *THIS - 4: Wend  
      If *THIS\__COM_IsDeadObject
        __COM_DEBUG("CRITICAL: "+Name$+"() --> Called on a dead (allready freed) object!", #COM_DEBUG_Critical)   
      EndIf
      ProcedureReturn *THIS\__COM_IsDeadObject ; asm code needs it too
    EndProcedure
    
    DataSection
      !__COM_DEBUG_#ObjectName#_CheckDead#__COM_COLON
      Data.l @__COM_DEBUG_#ObjectName#_CheckDead()  
    EndDataSection
  CompilerEndIf
EndMacro


Macro __COM_VTABLE_DEBUG_SetMethodDebug(ObjectName, InterfaceName, MethodName, NbArguments)  
  Goto __COM_#ObjectName#_#InterfaceName#_#MethodName#_DEBUG_END
  __COM_#ObjectName#_#InterfaceName#_#MethodName#_DEBUG_START:  
  
    ; call dead method check if enabled
    CompilerIf Defined(__COM_DEADTRACKING_ENABLED, #PB_Constant) 
    !  PUSH __COM_#ObjectName#_#InterfaceName#_#MethodName#_DEBUG_NAME
    !  PUSH dword [esp+8]  ; push *THIS
    !  CALL dword [__COM_DEBUG_#ObjectName#_CheckDead]
    !  CMP  eax, 0 ; check if it was a dead object (if so, do not execute the method)   
    !  JE   __COM_#ObjectName#_#InterfaceName#_#MethodName#_NOT_DEAD
    !  MOV  eax, dword 2147942487  ; E_INVALIDARG
    !  RET  (NbArguments+1)*4      ; return directly to caller
    !__COM_#ObjectName#_#InterfaceName#_#MethodName#_NOT_DEAD#__COM_COLON
    CompilerEndIf  
    
    __COM_DEBUG("CALLING: "+__COM_STRING(ObjectName)+"::"+__COM_STRING(InterfaceName)+"::"+__COM_STRING(MethodName)+"()", #COM_DEBUG_MethodEntry)    
    !  POP  eax  ; duplicate returnvalue
    !  PUSH eax
    !  PUSH eax
    !  MOV  eax, NbArguments+1  ; move old returnvalue before all arguments
    !  MOV  edx, esp
    !  ADD  edx, 4  ; move away from duplicate returnvalue
    !__COM_#ObjectName#_#InterfaceName#_#MethodName#_LOOP#__COM_COLON ; loop runs atleast once because of *THIS
    !  MOV  ecx, dword [edx+4]
    !  MOV  [edx], ecx
    !  ADD  edx, 4
    !  SUB  eax, 1
    !  CMP  eax, 0
    !  JNE  __COM_#ObjectName#_#InterfaceName#_#MethodName#_LOOP
    !  POP  eax  ; returnvalue
    !  MOV  [esp+NbArguments*4+4], eax
    !  POP  eax  ; get *THIS
    !  PUSH eax            
    !  MOV  eax, [eax]    
    !  ADD  eax, dword [__COM_#ObjectName#_#InterfaceName#_#MethodName#_DEBUG_OFFSET]
    !  CALL dword [eax] ; call real method
    !  PUSH eax
    !  PUSH __COM_#ObjectName#_#InterfaceName#_#MethodName#_DEBUG_NAME
    !  PUSH eax
    !  CALL dword [__COM_DEBUG_#ObjectName#_PrintResult]
    !  POP  eax
    !  RET
  __COM_#ObjectName#_#InterfaceName#_#MethodName#_DEBUG_END:
  DataSection
    !__COM_#ObjectName#_#InterfaceName#_#MethodName#_DEBUG_OFFSET#__COM_COLON    
    Data.l OffsetOf(__COM_#InterfaceName#_Vtbl_DEBUG\__COM_RealMethods)+OffsetOf(__COM_#InterfaceName#_Vtbl\MethodName)    
    !__COM_#ObjectName#_#InterfaceName#_#MethodName#_DEBUG_NAME#__COM_COLON
    Data.s __COM_STRING(ObjectName)+"::"+__COM_STRING(InterfaceName)+"::"+__COM_STRING(MethodName)
  EndDataSection    
  __COM_#ObjectName#_#InterfaceName#_Vtbl\__COM_RealMethods\MethodName = __COM_#ObjectName#_#InterfaceName#_Vtbl\MethodName
  __COM_#ObjectName#_#InterfaceName#_Vtbl\MethodName = ?__COM_#ObjectName#_#InterfaceName#_#MethodName#_DEBUG_START
EndMacro

Procedure __COM_CrashTraceMethod(*THIS)
  Protected index = 0
  
  CompilerIf #PB_Compiler_Thread
    LockMutex(__COM_DEBUG_ObjectListMutex)
  CompilerEndIf    
  
  While index < 9
    ForEach __COM_DEBUG_ObjectList()
      If __COM_DEBUG_ObjectList()\ObjectPointer = *THIS
        Break 2
      EndIf
    Next __COM_DEBUG_ObjectList()
    
    index + 1
    *THIS - 4
  Wend
  
  If index = 9
    __COM_DEBUG("CRITICAL: COM Method call outside of the VTable, on unknown Object/Interface!", #COM_DEBUG_Critical)
  Else
    __COM_DEBUG("CRITICAL: "+__COM_DEBUG_ObjectList()\ObjectNameString$+"::"+__COM_DEBUG_ObjectList()\InterfaceNameString$[index]+": COM Method call outside of VTable! (Passed the wrong Interface?)", #COM_DEBUG_Critical)
  EndIf
  
  CompilerIf #PB_Compiler_Thread
    UnlockMutex(__COM_DEBUG_ObjectListMutex)
  CompilerEndIf    
    
  ;
  ;
  ; ******************************************************************
  ; *                              NOTE:                             * 
  ; ******************************************************************
  ; *                                                                *
  ; *  If the PureBasic Debugger reports a crash at this line, then  *
  ; *  this means that somebody tried to call a Method that does     *
  ; *  exist on one of your Interfaces.                              *
  ; *  The debug message that was printed right before the crash     *
  ; *  should indicate the Object and Interface this happened with.  *
  ; *                                                                * 
  ; *  There is no error here in the framework code. This just       *
  ; *  happens to be the last line that could be successfully        *
  ; *  executed.                                                     *
  ; *                                                                *
  ; ******************************************************************
  ;
  ;  
EndProcedure
