;
; PropertyPage Shell Extension dll example
;   Adds a new page to the properties displayed for drives in the explorer.
;

; Usage:
;   To use it, you must build this source As a DLL And put the DLL in some known 
;   place (Not in C:\Windows\Temp Wink, so you remember where it is when you want 
;   To unregister it.
;
;   But you must register it first, With this DOS command:
;     C:\WINDOWS\system32>regsvr32 C:\PathToMyDLL\mydll.dll
;
;   To unregister it:
;     C:\WINDOWS\system32>regsvr32 /u C:\PathToMyDLL\mydll.dll
;
;
;   Then launch an explorer window And right click on a drive and select 'Properties'
;   There should be an extra empty page in the Properties called "Example Property Page"

; Try compiling the dll with debugging support enabled, and you can track all
; object use with a debugger like DebugView (http://www.sysinternals.com/Utilities/DebugView.html)
;

; Compile the dll with com debugging support enabled, and you can track all
; object use with a debugger like DebugView (http://www.sysinternals.com/Utilities/DebugView.html)
;
EnableCOMDebugging

; for the unload function 
Global TotalObjectCount, DllInstance

COMClass(PropertySheet)
  COMInterface(PropertySheet, IClassFactory)
  COMInterface(PropertySheet, IShellExtInit)
  COMInterface(PropertySheet, IShellPropSheetExt)   
  
  COMClassData(PropertySheet)
    Directory$     
    Thread.l
  EndCOMClassData  
  
  COMConstructor(PropertySheet)    
    TotalObjectCount + 1
    COMConstructorReturn IClassFactory
  EndCOMConstructor
  
  COMDestructor(PropertySheet)
    TotalObjectCount - 1
  EndCOMDestructor
EndCOMClass(PropertySheet, IClassFactory, IShellExtInit, IShellPropSheetExt)


#CLSID_MyPropertySheet = "{3b3dccdc-8263-4c16-8405-e1d58d6a5a46}"
#MyPropertySheetName  = "PBTestPropertySheet.1"

DefineCLSID(CLSID_MyPropertySheet, $3b3dccdc, $8263, $4c16, $84, $05, $e1, $d5, $8d, $6a, $5a, $46)



;- Method definitions
;
Procedure PropertySheet_IClassFactory_CreateInstance(*THIS.PropertySheet, *pUnkOuter.IUnknown, *IID.GUID, *pObject.LONG) COMMethodOf(PropertySheet)
  Protected i
  Protected IUnknown.IUnknown = *THIS

  If *pUnkOuter
    ProcedureReturn #CLASS_E_NOAGGREGATION
  Else
    ProcedureReturn IUnknown\QueryInterface(*IID, *pObject)
  EndIf
EndProcedure

COMEmptyMethod(PropertySheet, IClassFactory, LockServer, #E_FAIL)

Procedure PropertySheet_IShellExtInit_Initialize(*THIS.PropertySheet, *Folder.ITEMIDLIST, DataObject.IDataObject, hkeyProgID) COMMethodOf(PropertySheet)
  Protected Format.FORMATETC, Medium.STGMEDIUM
  Protected Result = #E_FAIL
  Protected Count, Length
  Protected init.INITCOMMONCONTROLSEX
  
  init\dwSize = SizeOf(INITCOMMONCONTROLSEX)
  init\dwICC  = #ICC_DATE_CLASSES|#ICC_STANDARD_CLASSES
  InitCommonControlsEx_(@init) ; must call this to avoid a crash when loading the dialog box
    
  
  ;
  ; NOTE: maybe implement the CFSTR_MOUNTEDVOLUME format to handle NTFS volumes
  ; mounted on a directory and not a driveletter
  ;
  
  If DataObject    
    Format\cfFormat = #CF_HDROP
    Format\ptd      = #Null
    Format\dwAspect = #DVASPECT_CONTENT
    Format\lindex   = -1
    Format\tymed    = #TYMED_HGLOBAL
    
    If DataObject\GetData(@Format, @Medium) = #S_OK
      Count = DragQueryFile_(Medium\hGlobal, $FFFFFFFF, #Null, 0)
      
      If Count = 1
        Length = DragQueryFile_(Medium\hGlobal, 0, #Null, 0)
        *THIS\Directory$ = Space(Length)
        DragQueryFile_(Medium\hGlobal, 0, @*THIS\Directory$, Length+1)    
        
        Result = #S_OK    
      EndIf

      ReleaseStgMedium_(@Medium)
    EndIf    
  EndIf

  ProcedureReturn Result
EndProcedure


Procedure PropertySheetPageProc(hWnd, uMsg, *page.PROPSHEETPAGE)
  Protected *THIS.PropertySheet = *page\lParam
  Protected ThisIUnknown.IUnknown = *page\lParam

  Select uMsg
  
    Case #PSPCB_ADDREF:  OutputDebugString_("Page is being added") 
      ThisIUnknown\AddRef() ; prevent the dll/object from being destroyed while the page exists
    
    Case #PSPCB_CREATE:  OutputDebugString_("Dialog box is being created")
    
    Case #PSPCB_RELEASE: OutputDebugString_("Page is being destroyed")
      ThisIUnknown\Release() ; allow release of the dll/object
      
  EndSelect

  ProcedureReturn #True
EndProcedure

Macro DialogReturn(Window, Value)
  SetWindowLong_(Window, #DWL_MSGRESULT, Value)
  ProcedureReturn #True
EndMacro

#NF_QUERY = 3
#NF_REQUERY = 4
#NFR_ANSI = 1
#NFR_UNICODE = 2

Procedure PropertySheetDialogProc(Window, Message, wParam, lParam)
  Protected *hdr.NMHDR

  Select Message
  
    Case #WM_INITDIALOG

    Case #WM_DESTROY

    Case #WM_NOTIFY
      *hdr.NMHDR = lParam
      OutputDebugString_("WM_NOTIFY: code = "+Str(*hdr\code))
      
      If *hdr\code = #PSN_SETACTIVE
        DialogReturn(Window, 0) ; important to handle this, otherwise we get a crash!      
      EndIf      
  
  EndSelect
  
  ProcedureReturn #False
EndProcedure

; This is a dialog template for an empty dialog
;
DataSection

  Dialog:
    Data.w 1 ; DlgVer
    Data.w $FFFF ; extended dlg structure
    Data.l 0 ; helpID
    Data.l 0 ; exStyle  
    Data.l #WS_CHILD | #WS_VISIBLE  ; Style    
    Data.w 0 ; Nb of items in the dialog    
    Data.w 0 ; x
    Data.w 0 ; y
    Data.w 0 ; w
    Data.w 0 ; h
    Data.w 0 ; PADDING
    Data.w $0000 ; NO menu!
    Data.w 0 ; PADDING
    Data.w $0000 ; No windowClass    
    Data.w 0 ; PADDING
    Data.w $0000 ; NO title
    Data.w 0 ; PADDING
        
EndDataSection

Prototype AddPropSheetPageProc(hPage, lParam)

Procedure PropertySheet_IShellPropSheetExt_AddPages(*THIS.PropertySheet, AddPage.AddPropSheetPageProc, lParam) COMMethodOf(PropertySheet)
  Protected page.PROPSHEETPAGE, hPage  
  Protected Result = #E_FAIL  

  page\dwSize      = SizeOf(PROPSHEETPAGE)
  page\dwFlags     = #PSP_USETITLE | #PSP_USECALLBACK | #PSP_DLGINDIRECT
  page\hInstance   = DllInstance
  page\pszTitle    = @"Example Property Page"
  page\pfnDlgProc  = @PropertySheetDialogProc()
  page\pfnCallback = @PropertySheetPageProc()
  page\lParam      = *THIS
  page\pszTemplate = ?Dialog   
  
  hPage = CreatePropertySheetPage_(@page)
  If hPage
    If AddPage(hPage, lParam)
      Result = #S_OK
    Else
      DestroyPropertySheetPage_(hPage)
    EndIf
  EndIf
  
  ProcedureReturn Result
EndProcedure


;- DLL Management
;
ProcedureDLL AttachProcess(Instance)
  DllInstance = Instance
 
  ;DisableThreadLibraryCalls_(Instance)
  
  OutputDebugString_("AttachProcess")
  
  BuildCOMVTable_IClassFactory(PropertySheet)
  BuildCOMVTable_IShellExtInit(PropertySheet)
  BuildCOMVTable_IShellPropSheetExt(PropertySheet)  
  
EndProcedure

ProcedureDLL DetachProcess(Instance)
  OutputDebugString_("DetachProcess")
EndProcedure

ProcedureDLL DllGetClassObject(*CLSID.CLSID, *IID.IID, *pObject.LONG)
  OutputDebugString_("DllGetClassObject")

  If *pObject And CompareCLSID(*CLSID, ?CLSID_MyPropertySheet) And CompareIID(*IID, ?IID_IClassFactory)
    *pObject\l = New_PropertySheet()
    ProcedureReturn #S_OK
  Else
    *pObject\l = 0
    ProcedureReturn #CLASS_E_CLASSNOTAVAILABLE  
  EndIf

EndProcedure

ProcedureDLL DllCanUnloadNow()  
  If TotalObjectCount = 0 
    OutputDebugString_("DllCanUnloadNow -> OK")
    ProcedureReturn #S_OK
  Else
    OutputDebugString_("DllCanUnloadNow -> FALSE")
    ProcedureReturn #S_FALSE
  EndIf
EndProcedure


; You can get the property page To add when a folder is selected.
; Add the CLSID under
; HKCR\Folder\ShellEx\PropertySheetHandlers
; And it will also work For folders



ProcedureDLL DllRegisterServer()
  OutputDebugString_("DllRegisterServer")
  Protected hKey1
  Protected DLLFileName$ = ProgramFilename()
  
  If RegCreateKeyEx_(#HKEY_CLASSES_ROOT, "Drive\shellex\PropertySheetHandlers\"+#MyPropertySheetName, 0, 0, #REG_OPTION_NON_VOLATILE, #KEY_ALL_ACCESS, #Null, @hKey1, 0) = #ERROR_SUCCESS
    RegSetValueEx_(hKey1, "", 0, #REG_SZ, #CLSID_MyPropertySheet, Len(#CLSID_MyPropertySheet)+1)
    RegCloseKey_(hKey1)

    If RegCreateKeyEx_(#HKEY_CLASSES_ROOT, #MyPropertySheetName+"\shellex\PropertySheetHandlers\MyPropertySheet.1", 0, 0, #REG_OPTION_NON_VOLATILE, #KEY_ALL_ACCESS, #Null, @hKey1, 0) = #ERROR_SUCCESS
      RegSetValueEx_(hKey1, "", 0, #REG_SZ, #CLSID_MyPropertySheet, Len(#CLSID_MyPropertySheet)+1)
      RegCloseKey_(hKey1)
      
      If RegCreateKeyEx_(#HKEY_LOCAL_MACHINE, "SOFTWARE\Microsoft\Windows\CurrentVersion\Shell Extensions\Approved", 0, 0, #REG_OPTION_NON_VOLATILE, #KEY_ALL_ACCESS, #Null, @hKey1, 0) = #ERROR_SUCCESS
        RegSetValueEx_(hKey1, #CLSID_MyPropertySheet, 0, #REG_SZ, #MyPropertySheetName, Len(#MyPropertySheetName)+1)
        RegCloseKey_(hKey1)
        
        If RegCreateKeyEx_(#HKEY_CLASSES_ROOT, "CLSID\"+#CLSID_MyPropertySheet, 0, 0, #REG_OPTION_NON_VOLATILE, #KEY_ALL_ACCESS, #Null, @hKey1, 0) = #ERROR_SUCCESS
          RegSetValueEx_(hKey1, "", 0, #REG_SZ, #MyPropertySheetName, Len(#MyPropertySheetName)+1)
          RegCloseKey_(hKey1)
          
          If RegCreateKeyEx_(#HKEY_CLASSES_ROOT, "CLSID\"+#CLSID_MyPropertySheet+"\InProcServer32", 0, 0, #REG_OPTION_NON_VOLATILE, #KEY_ALL_ACCESS, #Null, @hKey1, 0) = #ERROR_SUCCESS
            RegSetValueEx_(hKey1, "", 0, #REG_SZ, @DLLFileName$, Len(DLLFileName$)+1)
            RegSetValueEx_(hKey1, "ThreadingModel", 0, #REG_SZ, "Apartment", Len("Apartment")+1)          
            RegCloseKey_(hKey1)
            

            ProcedureReturn #S_OK ; success
          EndIf
        EndIf
      EndIf      
    EndIf
  EndIf
  
  ProcedureReturn #SELFREG_E_CLASS ; error
EndProcedure

ProcedureDLL DllUnregisterServer()
  OutputDebugString_("DllUnregisterServer")
  Protected hKey1
  
  If RegDeleteKey_(#HKEY_CLASSES_ROOT, "Drive\shellex\PropertySheetHandlers\"+#MyPropertySheetName) = #ERROR_SUCCESS
    If RegDeleteKey_(#HKEY_CLASSES_ROOT, #MyPropertySheetName+"\shellex\PropertySheetHandlers\MyPropertySheet1") = #ERROR_SUCCESS 
      If RegDeleteKey_(#HKEY_CLASSES_ROOT, #MyPropertySheetName+"\shellex\PropertySheetHandlers") = #ERROR_SUCCESS 
        If RegDeleteKey_(#HKEY_CLASSES_ROOT, #MyPropertySheetName+"\shellex") = #ERROR_SUCCESS
          If RegDeleteKey_(#HKEY_CLASSES_ROOT, #MyPropertySheetName) = #ERROR_SUCCESS        
          
            If RegOpenKeyEx_(#HKEY_LOCAL_MACHINE, "Software\Microsoft\Windows\CurrentVersion\Shell Extensions\Approved", 0, #KEY_ALL_ACCESS, @hKey1) = #ERROR_SUCCESS
              RegDeleteValue_(hKey1, #CLSID_MyPropertySheet)
              RegCloseKey_(hKey1)
              
              If RegOpenKeyEx_(#HKEY_CLASSES_ROOT, "CLSID\"+#CLSID_MyPropertySheet, 0, #KEY_ALL_ACCESS, @hKey1) = #ERROR_SUCCESS
                RegDeleteKey_(hKey1, "InProcServer32")
                RegCloseKey_(hKey1)
                
                If RegOpenKeyEx_(#HKEY_CLASSES_ROOT, "CLSID", 0, #KEY_ALL_ACCESS, @hKey1) = #ERROR_SUCCESS
                  RegDeleteKey_(hKey1, #CLSID_MyPropertySheet)
                  RegCloseKey_(hKey1)
                  
                  ProcedureReturn #S_OK ; success
                EndIf
              EndIf
            EndIf          
          EndIf
        EndIf
      EndIf
    EndIf
  EndIf
 
  ProcedureReturn #SELFREG_E_CLASS ; error
EndProcedure