;
; Example for creating a very basic IBindStatusCallback implementation
; to use with UrlDownloadToFile_()
;

; Run the example, then enable then enable the 'EnableCOMDebugging' line to see
; the generated debug output during execution. Also test the different debuglevels
; to see the different types of output levels.
;
; EnableCOMDebugging
; DebugLevel #COM_DEBUG_ObjectStats

; Define our class called 'StatusObject'
;
COMClass(StatusObject)
  ; Implement IBindStatusCallback only. It extends no other interface (except IUnknown)
  COMInterface(StatusObject, IBindStatusCallback) 
  
  ; We implement a constructor only to directly return a IBindStatusCallback
  ; pointer instead of the default IUnknown, as that is what UrlDownloadToFile_() expects.  
  COMConstructor(StatusObject)
    COMConstructorReturn IBindStatusCallback
  EndCOMConstructor  
  
  ; finish the definition. All implemented interfaces must be listed here once again.
EndCOMClass(StatusObject, IBindStatusCallback)

; For URLDownloadToFile, we really only need to implement the OnProgress() method
; The rest we leave out. The framework will automatically insert default methods
; which return #E_NOTIMPL for the other methods.
;
Procedure StatusObject_IBindStatusCallback_OnProgress(*THIS.StatusObject, Progress.l, ProgressMax.l, StatusCode.l, szStatusText.l) COMMethodOf(StatusObject)
  Protected percent.f

  If ProgressMax = 0
    percent.f = 0
  Else
    percent.f = (Progress*100)/ProgressMax
  EndIf
    
  Debug Str(Progress)+"/"+Str(ProgressMax)+"  ("+StrF(percent, 1)+"%)"

  ProcedureReturn #S_OK
EndProcedure

; This is important to build the IBindStatusCallback VTable. Otherwise accessing the
; object will lead to crashes. It is important that this call is placed somewhere
; where it is executed before the first object is used!
;
BuildCOMVTable_IBindStatusCallback(StatusObject)

; ---------------------------------------------------------


Define RemoteName$, LocalName$, StatusObject.IUnknown, BindStatus.IBindStatusCallback

; get user input
RemoteName$ = InputRequester("File Download:", "Enter remote FileName: ", "http://www.purebasic.com/")
If RemoteName$ = "": End: EndIf

LocalName$ = SaveFileRequester("Select Local Filename:", "C:\temp.html", "All Files|*.*", 1)
If LocalName$ = "": End: EndIf

; this creates a new instance of the object. Usually we get IUnknown here,
; but because of the COMConstructorReturn in the constructor, we get IBindStatusCallback directly.
;
BindStatus.IBindStatusCallback = New_StatusObject()

Debug "Starting download..."

; call the API with our object
;
If URLDownloadToFile_(0, @RemoteName$, @LocalName$, 0, BindStatus) = #S_OK
  Debug "Download successful."
Else
  Debug "Download failed."
EndIf

; release our reference to the object. Since the API has returned (and freed all its
; references to the object, it will be freed here)
;
BindStatus\Release()

End
